/**
@file       kisan_gpio_handler.c
@brief      AM5729 S2 B/D Character Device Driver
@details    GPIO Handler (/dev/kisan_gpio_handler)
@date       2020-07-10
@author     hong.yongje@kisane.com
*/

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>

#include <linux/gpio.h>          // gpio_set/get_value

#include <linux/miscdevice.h>           // misc_register
#include <linux/input.h>                // system call

#include <linux/uaccess.h>              // copy_to_user, copy_from_user
#include <asm/io.h>              // ioremap

#include <linux/delay.h>            // mdelay

#include <kisan/kisan_common.h>



#define DBG_MSG_ON_KISAN_GPIO_HANDLER

static void __iomem *g_ptrGpioBase2;
static void __iomem *g_ptrGpioBase3;
static void __iomem *g_ptrGpioBase4;
static void __iomem *g_ptrGpioBase5;
static void __iomem *g_ptrGpioBase6;
static void __iomem *g_ptrGpioBase7;
static void __iomem *g_ptrGpioBase8;


static int g_nDbgMsgFlag = __OFF__;
static int g_nToggleReset = __OFF__;

struct structGpioHandler {
    int nCmdType;   // 0=SetDirection, 1=GetDirection, 2=SetValue, 3=GetValue, 4=TogglePin
    int nGpioBank;
    int nGpioNum;
    int nCmdData;
    int nDbgMsgFlag;
};

typedef enum {
    CMD_TYPE_GPIO_HANDLER_SET_DIRECTION = 0,
    CMD_TYPE_GPIO_HANDLER_GET_DIRECTION,
    CMD_TYPE_GPIO_HANDLER_SET_VALUE,
    CMD_TYPE_GPIO_HANDLER_GET_VALUE,
    CMD_TYPE_GPIO_HANDLER_TOGGLE_PIN,
    MAX_NUM_OF_CMD_TYPE_GPIO_HANDLER
} eCommandTypeGpioHandler;


extern void BitsDisplay(int nVal, int nBitLen);


static int GetGpioDirection(int nGpioBank, int nGpioNum)
{
    unsigned int unRegVal = 0x00000000;
    int nRet = 0;

    void __iomem *iomem = NULL;

    if( (nGpioBank <= 1) || (nGpioBank > 8) ) {
        printk(KERN_ERR "[%s:%4d:%s] Wrong Gpio Bank (%d) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioBank);
        return -1;
    }

    if( (nGpioNum < 0) || (nGpioNum > 31) ) {
        printk(KERN_ERR "[%s:%4d:%s] Wrong Gpio Num (%d) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioNum);
        return -1;
    }

    switch(nGpioBank) {
        case 2:    iomem = g_ptrGpioBase2;    break;
        case 3:    iomem = g_ptrGpioBase3;    break;
		case 4:    iomem = g_ptrGpioBase4;	  break;
		case 5:    iomem = g_ptrGpioBase5;	  break;
		case 6:    iomem = g_ptrGpioBase6;	  break;
		case 7:    iomem = g_ptrGpioBase7;	  break;
		case 8:    iomem = g_ptrGpioBase8;	  break;
    }

    unRegVal = ioread32 (iomem + GPIO_OE);
    nRet = BIT_CHECK(unRegVal, nGpioNum);

    if(g_nDbgMsgFlag == __ON__) {
        printk(KERN_DEBUG "[%s:%4d:%s] Read GPIO%d_%d direction: %d (0=Out, 1=In) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioBank, nGpioNum, nRet);
        BitsDisplay(unRegVal, 32);
    }

    return nRet;
}


static int SetGpioDirection(int nGpioBank, int nGpioNum, int nDirection)
{
    unsigned int unRegVal = 0x00000000;
    void __iomem *iomem = NULL;
	int i = 0;

    if( (nGpioBank <= 1) || (nGpioBank > 8) ) {
        printk(KERN_ERR "[%s:%4d:%s] Wrong Gpio Bank (%d) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioBank);
        return -1;
    }

    if( (nGpioNum < 0) || (nGpioNum > 31) ) {
        printk(KERN_ERR "[%s:%4d:%s] Wrong Gpio Num (%d) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioNum);
        return -1;
    }

    if( (nDirection != __IN__) && (nDirection != __OUT__) ) {
        printk(KERN_ERR "[%s:%4d:%s] Wrong Gpio direction (%d) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nDirection);
        return -1;
    }

    switch(nGpioBank) {
        case 2:    iomem = g_ptrGpioBase2;    break;
        case 3:    iomem = g_ptrGpioBase3;    break;
		case 4:    iomem = g_ptrGpioBase4;	  break;
		case 5:    iomem = g_ptrGpioBase5;	  break;
		case 6:    iomem = g_ptrGpioBase6;	  break;
		case 7:    iomem = g_ptrGpioBase7;	  break;
		case 8:    iomem = g_ptrGpioBase8;	  break;
    }

    unRegVal = ioread32 (iomem + GPIO_OE);

    //unRegVal |= (nDirection << nGpioNum);
    if(nDirection == __OUT__)
	{
        BIT_CLR(unRegVal, nGpioNum);
	    iowrite32 (unRegVal, iomem + GPIO_OE);
		
		for(i = 0 ; i < 2 ; i++)
		{
//			unRegVal = ioread32 (iomem + GPIO_IRQWAKEN(i));
//			BIT_CLR(unRegVal, nGpioNum);
//			iowrite32 (unRegVal, iomem + GPIO_IRQWAKEN(i)); 
			
			unRegVal = ioread32 (iomem + GPIO_IRQSTATUS_SET(i));
			BIT_CLR(unRegVal, nGpioNum);
			iowrite32 (unRegVal, iomem + GPIO_IRQSTATUS_SET(i)); 
		}
	}
    else if(nDirection == __IN__)
	{
        BIT_SET(unRegVal, nGpioNum);
	    iowrite32 (unRegVal, iomem + GPIO_OE); 

		for(i = 0 ; i < 2 ; i++)
		{
//		    unRegVal = ioread32 (iomem + GPIO_IRQWAKEN(i));
//	        BIT_SET(unRegVal, nGpioNum);
//		    iowrite32 (unRegVal, iomem + GPIO_IRQWAKEN(i)); 
			
			unRegVal = ioread32 (iomem + GPIO_IRQSTATUS_SET(i));
			BIT_SET(unRegVal, nGpioNum);
			iowrite32 (unRegVal, iomem + GPIO_IRQSTATUS_SET(i)); 
		}
	}


	if(g_nDbgMsgFlag == __ON__) {
		printk(KERN_DEBUG "[%s:%4d:%s] GPIO%d_%d set direction to %d (0=Out, 1=In) \n",
			__FILENAME__, __LINE__, __FUNCTION__, nGpioBank, nGpioNum, nDirection);
		BitsDisplay(unRegVal, 32);
	}
	else {
		if(g_nToggleReset == __ON__)
		{
			mdelay(2);
		}
	}


    return 0;
}


static int GetGpioValue(int nGpioBank, int nGpioNum)
{
    unsigned int unRegVal = 0x00000000;
    void __iomem *iomem = NULL;

    int nRet = 0;

    if( (nGpioBank <= 1) || (nGpioBank > 8) ) {
        printk(KERN_ERR "[%s:%4d:%s] Wrong Gpio Bank (%d) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioBank);
        return -1;
    }

    if( (nGpioNum < 0) || (nGpioNum > 31) ) {
        printk(KERN_ERR "[%s:%4d:%s] Wrong Gpio Num (%d) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioNum);
        return -1;
    }

#if 1
    switch(nGpioBank) {
        case 2:    iomem = g_ptrGpioBase2;    break;
        case 3:    iomem = g_ptrGpioBase3;    break;
		case 4:    iomem = g_ptrGpioBase4;	  break;
		case 5:    iomem = g_ptrGpioBase5;	  break;
		case 6:    iomem = g_ptrGpioBase6;	  break;
		case 7:    iomem = g_ptrGpioBase7;	  break;
		case 8:    iomem = g_ptrGpioBase8;	  break;
    }

	unRegVal = ioread32(iomem + GPIO_OE);

    if( (BIT_CHECK(unRegVal, nGpioNum)) != __OUT__ ) {
        printk(KERN_DEBUG "[%s:%4d:%s] GPIO%d_%d direction is not out, can not set value. \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioBank, nGpioNum);
        return -1;
    }




	
	if( (BIT_CHECK(unRegVal, nGpioNum)) == __OUT__ )
	{
	    unRegVal = ioread32(iomem + GPIO_DATAOUT);
	}
	else
	{
	    unRegVal = ioread32(iomem + GPIO_DATAIN);
	}
	

    if(BIT_CHECK(unRegVal, nGpioNum))
        nRet = __HI__;
    else
        nRet = __LO__;

    if(g_nDbgMsgFlag == __ON__) {
        printk(KERN_DEBUG "[%s:%4d:%s] Get GPIO%d_%d value < %s \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioBank, nGpioNum, (nRet == __LO__ ? "Low(0)" : "Hi(1)"));
        BitsDisplay(unRegVal, 32);
    }

#else
	// API   ͸  ϴ Ϳ  
    nRet = gpio_get_value((32*nGpioBank)+nGpioNum);
#endif

    return nRet;
}


static int SetGpioValue(int nGpioBank, int nGpioNum, int nLowHi)
{
#if 1
    unsigned int unRegVal = 0x00000000;
    void __iomem *iomem = NULL;

    if( (nGpioBank <= 1) || (nGpioBank > 8) ) {
        printk(KERN_ERR "[%s:%4d:%s] Wrong Gpio Bank (%d) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioBank);
        return -1;
    }

    if( (nGpioNum < 0) || (nGpioNum > 31) ) {
        printk(KERN_ERR "[%s:%4d:%s] Wrong Gpio Num (%d) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioNum);
        return -1;
    }

    if( (nLowHi != __LO__) && (nLowHi != __HI__) ) {
        printk(KERN_ERR "[%s:%4d:%s] Wrong Gpio Value (%d) \n",
            __FILENAME__, __LINE__, __FUNCTION__, nLowHi);
        return -1;
    }

    switch(nGpioBank) {
        case 2:    iomem = g_ptrGpioBase2;    break;
        case 3:    iomem = g_ptrGpioBase3;    break;
		case 4:    iomem = g_ptrGpioBase4;	  break;
		case 5:    iomem = g_ptrGpioBase5;	  break;
		case 6:    iomem = g_ptrGpioBase6;	  break;
		case 7:    iomem = g_ptrGpioBase7;	  break;
		case 8:    iomem = g_ptrGpioBase8;	  break;
    }

	// TRM Pg. 7103

    unRegVal = ioread32(iomem + GPIO_OE);
    if( (BIT_CHECK(unRegVal, nGpioNum)) != __OUT__ ) {
        printk(KERN_DEBUG "[%s:%4d:%s] GPIO%d_%d direction is not out, can not set value. \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioBank, nGpioNum);
        return -1;
    }

    unRegVal = ioread32(iomem + GPIO_DATAOUT);

    //unRegVal |= (nLowHi << nGpioNum);
    if(nLowHi == __LO__)
        BIT_CLR(unRegVal, nGpioNum);
    else if(nLowHi == __HI__)
        BIT_SET(unRegVal, nGpioNum);

    iowrite32(unRegVal, iomem + GPIO_DATAOUT);

    if(g_nDbgMsgFlag == __ON__) {
        printk(KERN_DEBUG "[%s:%4d:%s] GPIO%d_%d set value > %s *** \n",
            __FILENAME__, __LINE__, __FUNCTION__, nGpioBank, nGpioNum, (nLowHi == __LO__ ? "Low(0)" : "Hi(1)"));
        BitsDisplay(unRegVal, 32);
    }
    else {        
        if(g_nToggleReset == __ON__)
        {
            mdelay(2);
        }
    }
#else
	// API   ͸  ϴ Ϳ  
    gpio_set_value(((32*nGpioBank)+nGpioNum), nLowHi);
#endif

    return 0;
}


static int ToggleGpioPin(int nGpioBank, int nGpioNum, int nTime)
{
    int nRet = 0;
    int nVal = 0;
    unsigned long ulEndTime = 0;

    int nIsDirectionChanged = __NO__;

    if(GetGpioDirection(nGpioBank, nGpioNum) != __OUT__) {
        SetGpioDirection(nGpioBank, nGpioNum, __OUT__);
        nIsDirectionChanged = __YES__;
    }

    nVal = GetGpioValue(nGpioBank, nGpioNum);
//    nVal = gpio_get_value((32*nGpioBank)+nGpioNum);

    switch(nVal)
    {
        case __LO__:        nRet = SetGpioValue(nGpioBank, nGpioNum, __HI__);        break;
        case __HI__:        nRet = SetGpioValue(nGpioBank, nGpioNum, __LO__);        break;
//        case __LO__:        gpio_set_value(((32*nGpioBank)+nGpioNum), __HI__);       break;
//        case __HI__:        gpio_set_value(((32*nGpioBank)+nGpioNum), __LO__);       break;

        default:
            printk(KERN_ERR "[%s:%4d:%s] Wrong GPIO%d_%d value: %d \n",
                __FILENAME__, __LINE__, __FUNCTION__, nGpioBank, nGpioNum, nVal);
            nRet = -1;
            break;
    }

    if(nRet >= 0 && nTime > 0)
	{
#if 0
        if( (nTime > 0) && (nTime < 1000) ) {
            udelay(nTime);
        }
        else if( (nTime >= 1000) && (nTime <= 5000) ) {
            mdelay(nTime / 1000);
        }
        else {
            printk(KERN_ERR "[%s:%4d:%s] Wrong GPIO%d_%d toggle delay time: %d \n",
                __FILENAME__, __LINE__, __FUNCTION__, nGpioBank, nGpioNum, nTime);
        }
#else
        // Delay N ms
        ulEndTime = get_jiffies_64() + ((HZ/1000)*nTime);
        while(jiffies < ulEndTime);
#endif

        nRet = SetGpioValue(nGpioBank, nGpioNum, nVal);
//        gpio_set_value(((32*nGpioBank)+nGpioNum), nVal);
    }

    if(nIsDirectionChanged == __YES__)
        SetGpioDirection(nGpioBank, nGpioNum, __IN__);

    return nRet;
}


static int InitGpioRegisterValues(void)
{
    void __iomem *cm_per;

    /*
	 TRM Pg. 1072, 1073, 1074, 1075

	 Table 3-1113. CM_L4PER_GPIO2_CLKCTRL: Offset(0x60), Physical Address(0x4A00_9760)
	 Table 3-1115. CM_L4PER_GPIO3_CLKCTRL: Offset(0x68), Physical Address(0x4A00_9768)
	 Table 3-1117. CM_L4PER_GPIO4_CLKCTRL: Offset(0x70), Physical Address(0x4A00_9770)
	 Table 3-1119. CM_L4PER_GPIO5_CLKCTRL: Offset(0x78), Physical Address(0x4A00_9778)
	 Table 3-1121. CM_L4PER_GPIO6_CLKCTRL: Offset(0x80), Physical Address(0x4A00_9770)
 	 Table 3-1155. CM_L4PER_GPIO7_CLKCTRL: Offset(0x110), Physical Address(0x4A00_9810)
	 Table 3-1157. CM_L4PER_GPIO8_CLKCTRL: Offset(0x118), Physical Address(0x4A00_9818)
	 	 
     Enable GPIO 2~8 clock
    */
    cm_per = ioremap(CSL_MPU_L4PER_CM_CORE_REGS, CSL_MPU_L4PER_CM_CORE_SIZE);
    if(!cm_per) {
      printk (KERN_ERR "[%s:%4d:%s] Failed to remap memory for CM_PER.\n",
        __FILENAME__, __LINE__, __FUNCTION__);
      return -1;
    }
    iowrite32(0x02, cm_per + 0x60); // MODULEMODE:  0x0=DISABLED,  0x2=ENABLED
    iowrite32(0x02, cm_per + 0x68); // MODULEMODE:  0x0=DISABLED,  0x2=ENABLED
    iowrite32(0x02, cm_per + 0x70); // MODULEMODE:  0x0=DISABLED,  0x2=ENABLED
    iowrite32(0x02, cm_per + 0x78); // MODULEMODE:  0x0=DISABLED,  0x2=ENABLED
    iowrite32(0x02, cm_per + 0x80); // MODULEMODE:  0x0=DISABLED,  0x2=ENABLED
    iowrite32(0x02, cm_per + 0x110); // MODULEMODE:  0x0=DISABLED,  0x2=ENABLED
    iowrite32(0x02, cm_per + 0x118); // MODULEMODE:  0x0=DISABLED,  0x2=ENABLED

    iounmap(cm_per);

	

    /*	 
 	 TRM Pg. 7109 ~ 7111
    */
    g_ptrGpioBase2 = ioremap(CSL_MPU_GPIO2_REGS, CSL_MPU_GPIO2_SIZE);
    if(!g_ptrGpioBase2) {
        printk (KERN_ERR "[%s:%4d:%s] Failed to ioremap g_ptrGpioBase2.\n",
            __FILENAME__, __LINE__, __FUNCTION__);
        return -1;
    }
    g_ptrGpioBase3 = ioremap(CSL_MPU_GPIO3_REGS, CSL_MPU_GPIO3_SIZE);
    if(!g_ptrGpioBase3) {
        printk (KERN_ERR "[%s:%4d:%s] Failed to ioremap g_ptrGpioBase3.\n",
            __FILENAME__, __LINE__, __FUNCTION__);
        return -1;
    }
    g_ptrGpioBase4 = ioremap(CSL_MPU_GPIO4_REGS, CSL_MPU_GPIO4_SIZE);
    if(!g_ptrGpioBase4) {
        printk (KERN_ERR "[%s:%4d:%s] Failed to ioremap g_ptrGpioBase4.\n",
            __FILENAME__, __LINE__, __FUNCTION__);
        return -1;
    }
    g_ptrGpioBase5 = ioremap(CSL_MPU_GPIO5_REGS, CSL_MPU_GPIO5_SIZE);
    if(!g_ptrGpioBase5) {
        printk (KERN_ERR "[%s:%4d:%s] Failed to ioremap g_ptrGpioBase5.\n",
            __FILENAME__, __LINE__, __FUNCTION__);
        return -1;
    }
    g_ptrGpioBase6 = ioremap(CSL_MPU_GPIO6_REGS, CSL_MPU_GPIO6_SIZE);
    if(!g_ptrGpioBase6) {
        printk (KERN_ERR "[%s:%4d:%s] Failed to ioremap g_ptrGpioBase6.\n",
            __FILENAME__, __LINE__, __FUNCTION__);
        return -1;
    }
    g_ptrGpioBase7 = ioremap(CSL_MPU_GPIO7_REGS, CSL_MPU_GPIO7_SIZE);
    if(!g_ptrGpioBase7) {
        printk (KERN_ERR "[%s:%4d:%s] Failed to ioremap g_ptrGpioBase7.\n",
            __FILENAME__, __LINE__, __FUNCTION__);
        return -1;
    }
    g_ptrGpioBase8 = ioremap(CSL_MPU_GPIO8_REGS, CSL_MPU_GPIO8_SIZE);
    if(!g_ptrGpioBase8) {
        printk (KERN_ERR "[%s:%4d:%s] Failed to ioremap g_ptrGpioBase8.\n",
            __FILENAME__, __LINE__, __FUNCTION__);
        return -1;
    } 


    return 0;
}


static int kisan_gpio_handler_open(struct inode *inode, struct file *file)
{
#if defined(DBG_MSG_ON_KISAN_GPIO_HANDLER)
//    printk(KERN_DEBUG "[%s:%4d:%s] \n", __FILENAME__, __LINE__, __FUNCTION__);
#endif  // #if defined(DBG_MSG_ON_KISAN_GPIO_HANDLER)

    return 0;
}


static ssize_t kisan_gpio_handler_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
#if defined(DBG_MSG_ON_KISAN_GPIO_HANDLER)
    printk(KERN_DEBUG "[%s:%4d:%s] \n", __FILENAME__, __LINE__, __FUNCTION__);
#endif  // #if defined(DBG_MSG_ON_KISAN_GPIO_HANDLER)

    return 0;
} 


static ssize_t kisan_gpio_handler_write(struct file *filp, const char __user *buf,
		size_t count, loff_t *f_pos)
{
#if defined(DBG_MSG_ON_KISAN_GPIO_HANDLER)
    printk(KERN_DEBUG "[%s:%4d:%s] \n", __FILENAME__, __LINE__, __FUNCTION__);
#endif  // #if defined(DBG_MSG_ON_KISAN_GPIO_HANDLER)

	return 0;
}


static long kisan_gpio_handler_ioctl(struct file *filp, unsigned int cmd, unsigned long args)
{
    void __user *argp = (void __user *) args;
    struct structGpioHandler stGpioHandler;

    int nCmdType = 0;   // 0= SetDirection, 1=GetDirection, 2=SetValue, 3=GetValue

    if(copy_from_user(&stGpioHandler, argp, sizeof(struct structGpioHandler))) {
        printk(KERN_ERR "[%s:%4d:%s] copy_from_user \n", __FILENAME__, __LINE__, __FUNCTION__);
        return -EFAULT;
    }

    nCmdType = stGpioHandler.nCmdType;
    g_nDbgMsgFlag = stGpioHandler.nDbgMsgFlag;
//    if(g_nDbgMsgFlag == __ON__) {
//        printk(KERN_DEBUG "[%s:%4d:%s] DbgMsg is ON. \n", __FILENAME__, __LINE__, __FUNCTION__);
//    }

    // Reset Flag For CPU Toggle
    g_nToggleReset = __OFF__;

    switch(nCmdType)
    {
        case CMD_TYPE_GPIO_HANDLER_SET_DIRECTION:
            if( (SetGpioDirection(stGpioHandler.nGpioBank, stGpioHandler.nGpioNum, stGpioHandler.nCmdData)) < 0 )
                return -EFAULT;
            break;

        case CMD_TYPE_GPIO_HANDLER_GET_DIRECTION:
            stGpioHandler.nCmdData = GetGpioDirection(stGpioHandler.nGpioBank, stGpioHandler.nGpioNum);
            if(copy_to_user(argp, &stGpioHandler, sizeof(struct structGpioHandler)))
                return -EFAULT;
            break;

        case CMD_TYPE_GPIO_HANDLER_SET_VALUE:
            if( (SetGpioValue(stGpioHandler.nGpioBank, stGpioHandler.nGpioNum, stGpioHandler.nCmdData)) < 0 )
                return -EFAULT;
            break;

        case CMD_TYPE_GPIO_HANDLER_GET_VALUE:
            stGpioHandler.nCmdData = GetGpioValue(stGpioHandler.nGpioBank, stGpioHandler.nGpioNum);
            if(copy_to_user(argp, &stGpioHandler, sizeof(struct structGpioHandler)))
                return -EFAULT;
            break;

        case CMD_TYPE_GPIO_HANDLER_TOGGLE_PIN:
            if(((stGpioHandler.nGpioBank == 0) && (stGpioHandler.nGpioNum == 17))        // Reset GPIO FPGA
                || ((stGpioHandler.nGpioBank == 2) && (stGpioHandler.nGpioNum == 2))     // Reset GPIO DSP
                || ((stGpioHandler.nGpioBank == 2) && (stGpioHandler.nGpioNum == 3))     // Reset GPIO MECHA
                ) 
            {
                g_nToggleReset = __ON__;
            }
            
            if( (ToggleGpioPin(stGpioHandler.nGpioBank, stGpioHandler.nGpioNum, stGpioHandler.nCmdData)) < 0 ) {
#if defined( DBG_MSG_ON_KISAN_GPIO_HANDLER )
                printk(KERN_DEBUG "[%s:%4d:%s] %d %d %d\n", __FILENAME__, __LINE__, __FUNCTION__,
                    stGpioHandler.nGpioBank, stGpioHandler.nGpioNum, stGpioHandler.nCmdData);
#endif  // #if defined( DBG_MSG_ON_KISAN_GPIO_HANDLER )
                return -EFAULT;
            }
            break;

        default:
            printk(KERN_ERR "[%s:%4d:%s] Wrong CmdType (%d) \n",
                __FILENAME__, __LINE__, __FUNCTION__, nCmdType);
            return -EINVAL;
            break;
    }

    return 0;
}


static struct file_operations st_kisan_gpio_handler_fops = {
	.owner   = THIS_MODULE,
    .open    = kisan_gpio_handler_open,
    .read    = kisan_gpio_handler_read,
    .write   = kisan_gpio_handler_write,
    .unlocked_ioctl   = kisan_gpio_handler_ioctl,
};


static struct miscdevice st_kisan_gpio_handler_dev = {
    .minor  = KISAN_GPIO_HANDLER_MINOR,		// ./kernel_root/include/linux/miscdevice.h
    .name   = "kisan_gpio_handler",
    .fops   = &st_kisan_gpio_handler_fops
};


static int __init kisan_gpio_handler_init(void)
{
	int nErr = 0;

    if( (InitGpioRegisterValues()) < 0 ) {
        printk(KERN_ERR "[%s:%4d:%s] Failed to InitGpioRegisterValues \n",
            __FILENAME__, __LINE__, __FUNCTION__);
        return -ENOMEM;
    }

    nErr = misc_register(&st_kisan_gpio_handler_dev);

    if( nErr ) {
        printk(KERN_ERR "[%s:%4d:%s] kisan_gpio_handler_init misc_register failure... \n",
            __FILENAME__, __LINE__, __FUNCTION__);

        return -ENOMEM;
    }
    else {
        printk(KERN_DEBUG "[%s:%4d:%s] Register a kisan_gpio_handler char driver. \n",
            __FILENAME__, __LINE__, __FUNCTION__);
    }

	return 0;
}


static void __exit kisan_gpio_handler_exit(void)
{
#if defined(DBG_MSG_ON_KISAN_GPIO_HANDLER)
    printk(KERN_DEBUG "[%s:%4d:%s] \n", __FILENAME__, __LINE__, __FUNCTION__);
#endif  // #if defined(DBG_MSG_ON_KISAN_GPIO_HANDLER)

    iounmap(g_ptrGpioBase2);
    iounmap(g_ptrGpioBase3);
    iounmap(g_ptrGpioBase4);
    iounmap(g_ptrGpioBase5);
    iounmap(g_ptrGpioBase6);
    iounmap(g_ptrGpioBase7);
    iounmap(g_ptrGpioBase8);

	misc_deregister(&st_kisan_gpio_handler_dev);
}


module_init(kisan_gpio_handler_init);
module_exit(kisan_gpio_handler_exit);

MODULE_DESCRIPTION("AM5729 Kisan GPIO Handling Driver");
MODULE_AUTHOR("hong.yongje@kisane.com");
MODULE_LICENSE("Dual BSD/GPL");

